// speaker specific vowel recognition program for dsPIC30F4013
// (toy speech recognition program)
// http://www.cepstrum.co.jp/

#include <p30F4013.h>
#include <stdlib.h>
#include <dsp.h>

#define LAG 8
#define LEN 300

_FWDT(WDT_OFF);
_FGS(CODE_PROT_OFF);
_FOSC(CSW_FSCM_OFF & HS);
_FBORPOR(PBOR_OFF & PWRT_64 & MCLR_EN);

fractional aa[LAG];
fractional ii[LAG];
fractional uu[LAG];
fractional ee[LAG];
fractional oo[LAG];
fractional cor[LAG];
fractional x[LEN];
fractional diff[5];
float      pwr;

int main(void) {
  int        i, j, min_pos, sw, avr;
  int        *p1, *p2;
  fractional d, s;

  TRISF=0x00;         // PORT F is output

  ADPCFG=0xEFFF;      // analog input pin is RA12/RB12
  ADCON1=0x0100;      // singed integer
  ADCON2=0x0000;      // Vref is AVSS/AVDD
  ADCON3=0x0002;      // A/D internal RC clock
  ADCHS =0x000C;      // input pin is RA12/RB12 
  ADCSSL=0x0000;      // no scan
  ADCON1bits.ADON=1;  // A/D converter on

  // clear autocorrelation data array
  for (i=0; i<LAG; i=i+1) {
    aa[i]=0;
    ii[i]=0;
    uu[i]=0;
    ee[i]=0;
    oo[i]=0;
  }

  while (1) {         // forever loop

    // A/D, Fs=5.9kHz (at 20MHz clock)
    for (i=0; i<LEN; i=i+1) {
      ADCON1bits.SAMP=1;                          // start sampling
      asm("nop \n nop \n nop \n nop \n nop");     // wait
      asm("nop \n nop \n nop \n nop \n nop");     // wait
      ADCON1bits.SAMP=0;                          // start conversion
      while (!ADCON1bits.DONE);                   // wait until coversion end
      x[i]=ADCBUF0;                               // read A/D data
      for (j=0; j<70; j=j+1) {                    // depend on optimization
        asm("nop \n nop \n nop \n nop \n nop");   // wait
      }
    }
  
    // differentiation (remove DC offset and equlization)
    for (i=0; i<LEN-1; i=i+1) x[i]=x[i]-x[i+1];

    // calculate autocorrelation
    for (i=0; i<LAG; i=i+1) {
      p1=&x[0];
      p2=&x[i];
      cor[i]=VectorDotProduct(LEN-LAG, p1, p2);
    }
    pwr=Fract2Float(cor[0]);
    for (i=0; i<LAG; i=i+1) {
      cor[i]=Float2Fract(Fract2Float(cor[i])/pwr/2.0);
    }
    avr=0;
    for (i=0; i<LAG; i=i+1) avr=avr+cor[i]/LAG;
    for (i=0; i<LAG; i=i+1) cor[i]=cor[i]-avr;

    sw=PORTB&0x1F;
    PORTF=0;
    if (sw!=0) {
      // make vowel template
      PORTF=(~sw)&0x1F;
      if ((sw&0x01)!=0) for (i=0; i<LAG; i=i+1) aa[i]=cor[i];
      if ((sw&0x02)!=0) for (i=0; i<LAG; i=i+1) ii[i]=cor[i];
      if ((sw&0x04)!=0) for (i=0; i<LAG; i=i+1) uu[i]=cor[i];
      if ((sw&0x08)!=0) for (i=0; i<LAG; i=i+1) ee[i]=cor[i];
      if ((sw&0x10)!=0) for (i=0; i<LAG; i=i+1) oo[i]=cor[i];
      for (i=0; i<10000; i=i+1) {
        for (j=0; j<10; j=j+1) {
          asm("nop \n nop \n nop \n nop \n nop");   // wait
        }
      }
      PORTF=0;
      for (i=0; i<10000; i=i+1) {
        for (j=0; j<10; j=j+1) {
          asm("nop \n nop \n nop \n nop \n nop");   // wait
        }
      }
      PORTF=(~sw)&0x1F;
      for (i=0; i<10000; i=i+1) {
        for (j=0; j<10; j=j+1) {
          asm("nop \n nop \n nop \n nop \n nop");   // wait
        }
      }
    } else if (0.00002<pwr) {
      // calculate diffrence between template and correlation
      PORTF=sw;
      for (i=0; i<5; i=i+1) diff[i]=0;
      for (i=0; i<LAG; i=i+1) {
        diff[0]=diff[0]+abs(cor[i]-aa[i])/LAG;
        diff[1]=diff[1]+abs(cor[i]-ii[i])/LAG;
        diff[2]=diff[2]+abs(cor[i]-uu[i])/LAG;
        diff[3]=diff[3]+abs(cor[i]-ee[i])/LAG;
        diff[4]=diff[4]+abs(cor[i]-oo[i])/LAG;
      }
      // calculate max value (recognition)
      p1=&diff[0];
      VectorMin(5, p1, &min_pos);
      PORTF=1<<min_pos;
    }

  }

}






